const {exec, spawn,} = require('child_process')
const fs = require('fs')
const path = require('path')
const { allowedFiles } = require('./filterUtils')
const getNodeModulesPath = require('../../utils/getNodeModulesPath')
const { Logger } = require('../../utils/Logger')

async function isMergeCommit() {
    return new Promise((resolve, reject) => {
        exec('git rev-parse -q --verify MERGE_HEAD', (error,stderr,stdout) => {
            if(error){
                reject(error.toString().trim())
            }
            else if(stderr){
                resolve(stderr.trim())
            }
            else if(stdout){
                resolve(stdout.trim())
            }
        })
    })

}

async function getCurrentBranch(){
    return new Promise((resolve, reject) => {
        exec('git rev-parse --abbrev-ref HEAD', (error,stderr,stdout) => {
            if(error){
                reject('Error fetching current branch')
            }
            else if(stderr){
                resolve(stderr)
            }
        })
    })
}

async function getStagedFiles(){
    return new Promise((resolve, reject) => {
        exec("git diff --staged --name-only",(error,stderr,stdout) =>{
           if (error){
            if(error != null)
            reject("Couldn't fetch staged files")
           }
           else if(stderr){
            resolve(stderr.trim().split('\n'))
           }
           else if(stdout.trim() === ''){
            resolve(stdout.trim())
           }
        })
    })
    
}

async function findEslintErrors(file){
    let nodeModulesPathOfProject = `${getNodeModulesPath()}`
    let eslintExecutablePath = `${nodeModulesPathOfProject}/node_modules/eslint/bin/eslint.js`
    let eslintConfigurationFilePath = `${nodeModulesPathOfProject}/lint.config.js`
    let isEslintExecutablePresent = fs.existsSync(eslintExecutablePath) ? true : false
    let isNodeModulesPresent = fs.existsSync(nodeModulesPathOfProject)
    let isEslintConfigurationFilePresent = fs.existsSync(eslintConfigurationFilePath)

    // if(!isEslintConfigurationFilePresent){
    //     createEslintConfigFile()
    // }
    if(isNodeModulesPresent){
        if(isEslintExecutablePresent){
            return new Promise ((resolve, reject) => {
                exec(`npx --ignore-existing "${eslintExecutablePath}" --no-eslintrc -c "${eslintConfigurationFilePath}" --no-inline-config --resolve-plugins-relative-to="${nodeModulesPathOfProject}/node_modules" ${file}`,(error,stderr,stdout) => {
                    if(stderr){
                        resolve(stderr.trim().split('\n'))
                    }
                    else if(error){
                        console.log(error)
                        reject("Error executing eslint command",error)
                    }
                    else{
                        resolve([])
                    }
                })
            })
        }
        else{
            Logger.log(Logger.INFO_TYPE,'Eslint executable not found. make sure eslint plugin is installed')
        }
    }
    else{
        Logger.log(Logger.INFO_TYPE,'node_modules not found')
    }
}
async function calculateGitDiffForFile(branch_name,file){
    let gitDiffCommand = `git diff -U0 ${branch_name.trim()} ${file}`
    return new Promise ((resolve,reject) => {        
        exec(gitDiffCommand,(error, stderr) => {
            if(stderr){
                resolve(stderr.trim().split('\n'))
            }
            else if(error) {
                reject(error)
            }
        })
    })
}
async function preCommitHook() {
    console.log('working dir  : ',process.cwd())
    try{
        let isMerge = await isMergeCommit();
        Logger.log(Logger.INFO_TYPE,'Looks like you have merged. So skipping pre commit check');
        process.exit(0);
    }
    catch(error){
        let staged_files = []
        let current_branch = ''
        let hasEslintErrorsInChangedLines = false
        let areFilesStaged = false
        Logger.log(Logger.INFO_TYPE,'Executing pre commit hook...')
        try{
            current_branch = await getCurrentBranch()
        }
        catch{
            Logger.log(Logger.INFO_TYPE,"Error fetching current branch")
        }
        try{
            staged_files = await getStagedFiles()
            if(!staged_files.length == 0){
                areFilesStaged = true
                for (file in staged_files){
                    let currentFileName = staged_files[file]
                    let changedLinesArray = []
                    let eslintErrorsInChangedLines = []

                    if(path.extname(staged_files[file]) === '.js'){
                        try{
                            let eslintErrorsInFile= await findEslintErrors(staged_files[file])
                            if(staged_files[file] && typeof(staged_files[file]) == 'string'){
                                if(! eslintErrorsInFile.length == 0){
                                    //git diff is computed and stored in an array
                                    let git_diff = await calculateGitDiffForFile(current_branch,staged_files[file])
                                    changedLinesArray = git_diff.filter((line) => line.startsWith('@@'))
                                    
                                    let changedLinesStartArray = []
                                    let changedLinesEndArray = []
                                    
                                    //Calculating changed lines in a file and storing them in respective arrays
                                    for (number of changedLinesArray){
                                        let changesStartLine = parseInt(number.split(' ')[2].split(',')[0])
                                        changedLinesStartArray.push(changesStartLine)
                                        let changesEndLine = number.split(' ')[2].split(',')[1]
                                        if(changesEndLine === undefined){
                                            changedLinesEndArray.push(changesStartLine)
                                        }
                                        else{
                                            changedLinesEndArray.push(changesStartLine + parseInt(changesEndLine) - 1)
                                        }
                                    }
                                    for (let error=1; error < eslintErrorsInFile.length - 2; error++){
                                        //eslintErrorsInFile[error].trim() - 69:26  error => Do not hardcode content. Use I18N key instead  no-hardcoding/no-hardcoding, 
                                        //eslintErrorsInFile[error].trim().split(' ')[0] => 69:26
                                        //eslintErrorsInFile[error].trim().split(' ')[0].split(':')[0] => 69
                                        
                                        let eslintErrorLineNumber = eslintErrorsInFile[error].trim().split(' ')[0].split(':')[0]
        
                                        for(lineNumber in changedLinesStartArray){
                                            if(eslintErrorLineNumber >= changedLinesStartArray[lineNumber] && 
                                                eslintErrorLineNumber <= changedLinesEndArray[lineNumber]){
                                                    eslintErrorsInChangedLines.push(eslintErrorsInFile[error])
                                                }
                                        }
                                    }
                                    if(eslintErrorsInChangedLines.length > 0){
                                        Logger.log(Logger.INFO_TYPE,currentFileName)
                                        for(eslintError of eslintErrorsInChangedLines){
                                            Logger.log(Logger.FAILURE_TYPE,eslintError.trimEnd())
                                        }
                                        hasEslintErrorsInChangedLines = true
                                    }  
                                }
                            } 
                        }
                        catch(err){
                            Logger.log(Logger.FAILURE_TYPE,err)
                            Logger.log(Logger.FAILURE_TYPE,"Error in executing eslint command")
                        }
                        
                    }
                }
            }
            else if(staged_files.length === 0){
                Logger.log(Logger.INFO_TYPE,'No files have been staged. Stage your files before committing')
            }
        }
        catch{
            Logger.log(Logger.INFO_TYPE,'Error executing pre commit hook')
        }
        if(hasEslintErrorsInChangedLines){
            Logger.log(Logger.FAILURE_TYPE,`There are eslint errors present. So commit is aborted`)
            process.exit(1)
        }
        else if(!hasEslintErrorsInChangedLines && areFilesStaged){
            Logger.log(Logger.SUCCESS_TYPE,`Commit Successful`)
            process.exit(0)
        }
    }
}


module.exports = {
    preCommitHook
}